/*
 * Authconfig - authentication configuration program
 * Copyright (c) 1999, 2000 Red Hat, Inc.
 *
 * This program is licensed under the terms of the GPL.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <glib.h>
#include <libgen.h>
#include <libintl.h>
#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "shvar.h"
#include "authinfo.h"

#ifdef LOCAL_POLICIES
#include "localpol.h"
#endif

#define LOGIC_REQUIRED		"required"
#define LOGIC_REQUISITE		"requisite"
#define LOGIC_SUFFICIENT	"sufficient"
#define LOGIC_OPTIONAL		"optional"
#define LOGIC_IGNORE_UNKNOWN	"[default=ok user_unknown=ignore service_err=ignore system_err=ignore]"

/* Read hesiod setup.  Luckily, /etc/hesiod.conf is simple enough that shvar
 * can read it just fine. */
gboolean
authInfoReadHesiod(struct authInfoType *info)
{
	shvarFile *sv = NULL;
	char *tmp = NULL;

	sv = svNewFile(SYSCONFDIR "/hesiod.conf");
	if(sv == NULL) {
		return FALSE;
	}

	tmp = svGetValue(sv, "lhs");
	if(tmp != NULL) {
		info->hesiodLHS = g_strdup(tmp);
		free(tmp);
		tmp = info->hesiodLHS + strlen(info->hesiodLHS);
		while((tmp > info->hesiodLHS) && isspace(tmp[-1])) {
			tmp[-1] = '\0';
			tmp--;
		}
	}

	tmp = svGetValue(sv, "rhs");
	if(tmp != NULL) {
		info->hesiodRHS = g_strdup(tmp);
		free(tmp);
		tmp = info->hesiodRHS + strlen(info->hesiodRHS);
		while((tmp > info->hesiodRHS) && isspace(tmp[-1])) {
			tmp[-1] = '\0';
			tmp--;
		}
	}

	svWriteFile(sv, 0644);
	svCloseFile(sv);
	sv = NULL;
	return TRUE;
}

/* Read NIS setup from /etc/yp.conf. */
gboolean
authInfoReadNIS(struct authInfoType *info)
{
	FILE *fp = NULL;
	char buf[LINE_MAX], *p, *q;

	/* Read NIS setup. */
	fp = fopen(SYSCONFDIR "/yp.conf", "r");
	if(fp == NULL) {
		return FALSE;
	}

	memset(buf, '\0', sizeof(buf));
	while(fgets(buf, sizeof(buf) - 1, fp) != NULL) {
		p = buf + strlen(buf);

		/* Snip off the terminating junk. */
		while((p > buf) && (isspace(p[-1]) || (p[-1] == '\n'))) {
			p[-1] = '\0';
			p--;
		}

		/* Skip initial whitespace. */
		for(p = buf; (isspace(*p) && (*p != '\0')); p++);

		/* Is it a "ypserver" statement? */
		if(strncmp("ypserver", p, 8) == 0) {
			/* Skip intervening whitespace. */
			for(p += 8; (isspace(*p) && (*p != '\0')); p++);

			/* Save the server's name. */
			if(*p != '\0') {
				if(info->nisServer != NULL) {
					char *tmp = info->nisServer;
					info->nisServer = g_strdup_printf("%s,%s", tmp, p);
					g_free(tmp);
				} else {
					info->nisServer = g_strdup(p);
				}
			}

			memset(buf, '\0', sizeof(buf));
			continue;
		}

		/* It had better be a "domain" statement. */
		if(strncmp("domain", p, 6) == 0) {
			/* Skip intervening whitespace. */
			for(p += 6; (isspace(*p) && (*p != '\0')); p++);

			/* Save the domain's name.  To do that, find its end. */
			for(q = p; (!isspace(*q) && (*q != '\0')); q++);
			if(*p != '\0') {
				info->nisDomain = g_strndup(p, q - p);
			}

			/* Skip over some more whitespace. */
			for(p = q; (isspace(*p) && (*p != '\0')); p++);

			/* Is it "server"?  If not, assume "broadcast". */
			if(strncmp(p, "server", 6) == 0) {
				for(p += 6; (isspace(*p) && (*p != '\0')); p++);
				if(*p != '\0') {
					if(info->nisServer != NULL) {
						char *tmp = info->nisServer;
						info->nisServer = g_strdup_printf("%s,%s", p, tmp);
						g_free(tmp);
					} else {
						info->nisServer = g_strdup(p);
					}
				}
			}

			memset(buf, '\0', sizeof(buf));
			continue;
		}

		memset(buf, '\0', sizeof(buf));
	}

	fclose(fp);
	return TRUE;
}

/* Read LDAP setup from /etc/ldap.conf. */
gboolean
authInfoReadLDAP(struct authInfoType *info)
{
	FILE *fp = NULL;
	char buf[LINE_MAX], *p;

	/* Read NIS setup. */
	fp = fopen(SYSCONFDIR "/ldap.conf", "r");
	if(fp == NULL) {
		return FALSE;
	}

	memset(buf, '\0', sizeof(buf));
	while(fgets(buf, sizeof(buf) - 1, fp) != NULL) {
		p = buf + strlen(buf);

		/* Snip off the terminating junk. */
		while((p > buf) && (isspace(p[-1]) || (p[-1] == '\n'))) {
			p[-1] = '\0';
			p--;
		}

		/* Skip initial whitespace. */
		for(p = buf; (isspace(*p) && (*p != '\0')); p++);

		/* Is it a "base" statement? */
		if(strncmp("base", p, 4) == 0) {
			/* Skip intervening whitespace. */
			for(p += 4; (isspace(*p) && (*p != '\0')); p++);

			/* Save the base DN. */
			if(*p != '\0') {
				info->ldapBaseDN = g_strdup(p);
			}

			memset(buf, '\0', sizeof(buf));
			continue;
		}

		/* Is it a "host" statement? */
		if(strncmp("host", p, 4) == 0) {
			/* Skip intervening whitespace. */
			for(p += 4; (isspace(*p) && (*p != '\0')); p++);

			/* Save the host name or IP. */
			if(*p != '\0') {
				info->ldapServer = g_strdup(p);
			}

			memset(buf, '\0', sizeof(buf));
			continue;
		}

		/* Is it a "ssl" statement? */
		if(strncmp("ssl", p, 3) == 0) {
			/* Skip intervening whitespace. */
			for(p += 3; (isspace(*p) && (*p != '\0')); p++);

			info->enableLDAPS = (strncmp(p, "start_tls", 9) == 0);

			memset(buf, '\0', sizeof(buf));
			continue;
		}

		memset(buf, '\0', sizeof(buf));
	}

	fclose(fp);
	return TRUE;
}

/* Read Kerberos setup from /etc/krb5.conf. */
gboolean
authInfoReadKerberos(struct authInfoType *info)
{
	FILE *fp = NULL;
	char buf[LINE_MAX], *p, *q;
	char *section = NULL;
	char *subsection = NULL;

	fp = fopen(SYSCONFDIR "/krb5.conf", "r");
	if(fp == NULL) {
		return FALSE;
	}

	memset(buf, '\0', sizeof(buf));

	while(fgets(buf, sizeof(buf) - 1, fp) != NULL) {
		p = buf + strlen(buf);

		/* Snip off the terminating junk. */
		while((p > buf) && (isspace(p[-1]) || (p[-1] == '\n'))) {
			p[-1] = '\0';
			p--;
		}

		/* Skip initial whitespace. */
		for(p = buf; (isspace(*p) && (*p != '\0')); p++);

		/* If it's a new section, note which one we're "in". */
		if(p[0] == '[') {
			p++;
			for(q = p; ((*q != ']') && (*q != '\0')); q++);

			if(section != NULL) {
				g_free(section);
			}
			if(subsection != NULL) {
				g_free(subsection);
				subsection = NULL;
			}
			if(q - p > 0)  {
				section = g_strndup(p, q - p);
			}

			memset(buf, '\0', sizeof(buf));
			continue;
		}

		/* Check for the default realm keyword. */
		if(section != NULL)
		if(strcmp(section, "libdefaults") == 0)
		if(strncmp(p, "default_realm", 13) == 0) {
			/* Skip intervening whitespace and the equal sign. */
			for(p += 13;
			    ((isspace(*p) || (*p == '=')) && (*p != '\0'));
			    p++);

			/* Save the default realm. */
			if(*p != '\0') {
				info->kerberosRealm = g_strdup(p);
			}

			memset(buf, '\0', sizeof(buf));
			continue;
		}

		/* Check for the section about the current realm. */
		if(section != NULL)
		if(strcmp(section, "realms") == 0)
		if(subsection == NULL) {
			/* Read the name of the realm. */
			for(q = p; (!isspace(*q) && (*q != '\0')); q++);

			if(q - p > 0)  {
				subsection = g_strndup(p, q - p);
			}

			memset(buf, '\0', sizeof(buf));
			continue;
		}

		/* Check for the end of a realm section. */
		if(section != NULL)
		if(strcmp(section, "realms") == 0)
		if(subsection != NULL)
		if(strncmp(p, "}", 1) == 0) {
			if(subsection != NULL) {
				g_free(subsection);
				subsection = NULL;
			}

			memset(buf, '\0', sizeof(buf));
			continue;
		}

		/* Values within the current realm. */
		if(section != NULL)
		if(strcmp(section, "realms") == 0)
		if(subsection != NULL)
		if(info->kerberosRealm != NULL)
		if(strcmp(subsection, info->kerberosRealm) == 0) {
			char **target = NULL, *tmp;

			/* See if this is a key we care about. */
			if(strncmp(p, "kdc", 3) == 0) {
				target = &info->kerberosKDC;
				p += 3;
			}
			if(strncmp(p, "admin_server", 12) == 0) {
				target = &info->kerberosAdminServer;
				p += 12;
			}
			if(target == NULL) {
				memset(buf, '\0', sizeof(buf));
				continue;
			}

			/* Skip over the variable and the equal sign. */
			while((isspace(*p) || (*p == '=')) && (*p != '\0')) p++;

			/* Append if we need to, else make a fresh string. */
			if((*target != NULL) && (*p != '\0')) {
				tmp = g_malloc0(strlen(p) + strlen(*target) + 2);
				sprintf(tmp, "%s,%s", *target, p);
				g_free(*target);
				*target = tmp;
			}
			if((*target == NULL) && (*p != '\0')) {
				*target = g_strdup(p);
			}
			memset(buf, '\0', sizeof(buf));
			continue;
		}

		memset(buf, '\0', sizeof(buf));
	}

	fclose(fp);
	return TRUE;
}

/* Read NSS setup from /etc/nsswitch.conf. */
gboolean
authInfoReadNSS(struct authInfoType *info)
{
	FILE *fp = NULL;
	char buf[LINE_MAX], *p;
	char *nss_config = NULL;

	/* Read NIS setup. */
	fp = fopen(SYSCONFDIR "/nsswitch.conf", "r");
	if(fp == NULL) {
		return FALSE;
	}

	while(fgets(buf, sizeof(buf) - 1, fp) != NULL) {
		p = buf + strlen(buf);

		/* Snip off the terminating junk. */
		while((p > buf) && (isspace(p[-1]) || (p[-1] == '\n'))) {
			p[-1] = '\0';
			p--;
		}

		/* Skip initial whitespace. */
		for(p = buf; (isspace(*p) && (*p != '\0')); p++);

		if(strncmp("passwd:", buf, 7) == 0) {
			/* Skip the keyword and whitespace. */
			for(p += 7; (isspace(*p) && (*p != '\0')); p++);
			if(*p != '\0') {
				nss_config = g_strdup(p);
			}
		}
	}

	if(nss_config != NULL) {
		info->enableDB = (strstr(nss_config, "db") != NULL);
		info->enableHesiod = (strstr(nss_config, "hesiod") != NULL);
		info->enableLDAP = (strstr(nss_config, "ldap") != NULL);
		/* Don't be fooled by "nisplus". */
		for(p = nss_config; strstr(p, "nis") != NULL; p++) {
			info->enableNIS = ((strstr(p, "nis") != NULL) &&
					   ((strstr(p, "nis"))[3] != 'p'));
			if(info->enableNIS) {
				break;
			}
		}
		info->enableNIS3 = (strstr(nss_config, "nisplus") != NULL);
	}

	fclose(fp);
	return TRUE;
}

/* Read hints from the PAM control file. */
gboolean
authInfoReadPAM(struct authInfoType *authInfo)
{
	char ibuf[LINE_MAX];
	char module[PATH_MAX];
	char flags[PATH_MAX];
	char *p, *q, *stack;
	FILE *fp;
	struct stat st;
	shvarFile *sv = NULL;
	char *tmp = NULL;

	fp = fopen(SYSCONFDIR "/pam.d/" AUTH_PAM_SERVICE, "r");
	if(fp == NULL) {
		return FALSE;
	}

	while(fgets(ibuf, sizeof(ibuf), fp) != NULL) {
		memset(module, '\0', sizeof(module));
		memset(flags, '\0', sizeof(flags));

		q = ibuf + strlen(ibuf);
		while((q > ibuf) && ((q[-1] == '\n') || (q[-1] == '\r'))) {
			q[-1] = '\0';
			q--;
		}

		p = ibuf;
		for(q = p; !isspace(*q) && (*q != '\0'); q++); /* stack */
		stack = p;
		if((strncmp(stack, "auth", 4) != 0) &&
		   (strncmp(stack, "account", 7) != 0)) {
			continue;
		}

		for(p = q; isspace(*p) && (*p != '\0'); p++);
		for(q = p; !isspace(*q) && (*q != '\0'); q++); /* control */

		for(p = q; isspace(*p) && (*p != '\0'); p++);
		for(q = p; !isspace(*q) && (*q != '\0'); q++); /* module */
		if(q - p < sizeof(module)) {
			strncpy(module, p, q - p);
			if(strstr(module, "pam_krb5")) {
				authInfo->enableKerberos = TRUE;
				continue;
			}
			if(strstr(module, "pam_ldap")) {
				authInfo->enableLDAPAuth = TRUE;
				continue;
			}
#ifdef LOCAL_POLICIES
			if(strstr(module, "pam_stack")) {
				authInfo->enableLocal = TRUE;
				continue;
			}
#endif
		}

		for(p = q; isspace(*p) && (*p != '\0'); p++);
		for(q = p; !isspace(*q) && (*q != '\0'); q++); /* flags */
		if(q - p < sizeof(module)) {
			if(strncmp(stack, "auth", 4) == 0)
			if(strstr(module, "pam_unix") ||
			   strstr(module, "pam_pwdb")) {
				authInfo->enableMD5 =
					(strstr(p, "md5") != NULL);
				authInfo->enableShadow =
					stat("/etc/shadow", &st) == 0;
				authInfo->enableBigCrypt =
					(strstr(p, "bigcrypt") != NULL);
			}
			if(strncmp(stack, "account", 7) == 0)
			if(strstr(module, "pam_unix")) {
				authInfo->brokenShadow =
					(strstr(p, "broken_shadow") != NULL);
			}
		}
	}

	fclose(fp);

	if(stat(SYSCONFDIR "/sysconfig/authconfig", &st) == 0) {
		sv = svNewFile(SYSCONFDIR "/sysconfig/authconfig");
	}
	if(sv != NULL) {
		tmp = svGetValue(sv, "USEMD5");
		if(tmp != NULL) {
			if(strcmp(tmp, "yes") == 0) {
				authInfo->enableMD5 = TRUE;
			}
			if(strcmp(tmp, "no") == 0) {
				authInfo->enableMD5 = FALSE;
			}
			free(tmp);
		}
		tmp = svGetValue(sv, "USEKERBEROS");
		if(tmp != NULL) {
			if(strcmp(tmp, "yes") == 0) {
				authInfo->enableKerberos = TRUE;
			}
			if(strcmp(tmp, "no") == 0) {
				authInfo->enableKerberos = FALSE;
			}
			free(tmp);
		}
		tmp = svGetValue(sv, "USELDAPAUTH");
		if(tmp != NULL) {
			if(strcmp(tmp, "yes") == 0) {
				authInfo->enableLDAPAuth = TRUE;
			}
			if(strcmp(tmp, "no") == 0) {
				authInfo->enableLDAPAuth = FALSE;
			}
			free(tmp);
		}
		svCloseFile(sv);
		sv = NULL;
	}

	return TRUE;
}

/* Read hints from the PAM control file. */
gboolean
authInfoReadNetwork(struct authInfoType *authInfo)
{
	shvarFile *sv = NULL;
	char *tmp = NULL;

	sv = svNewFile(SYSCONFDIR "/sysconfig/network");
	if(sv == NULL) {
		return FALSE;
	}

	if((tmp = svGetValue(sv, "NISDOMAIN")) != NULL) {
		if(authInfo->nisDomain) g_free(authInfo->nisDomain);
		authInfo->nisDomain = g_strdup(tmp);
		free(tmp);
	}

	svCloseFile(sv);

	return TRUE;
}

struct authInfoType *
authInfoRead()
{
	struct authInfoType *ret = NULL;

	ret = g_malloc0(sizeof(struct authInfoType));

	authInfoReadHesiod(ret);
	authInfoReadNIS(ret);
	authInfoReadLDAP(ret);
	authInfoReadKerberos(ret);
	authInfoReadNSS(ret);
	authInfoReadPAM(ret);
	authInfoReadNetwork(ret);

	return ret;
}

struct authInfoType *
authInfoCopy(struct authInfoType *info)
{
	struct authInfoType *ret = NULL;

	ret = g_malloc0(sizeof(struct authInfoType));

	*ret = *info;

        ret->hesiodLHS = info->hesiodLHS ? g_strdup(info->hesiodLHS) : NULL;
        ret->hesiodRHS = info->hesiodLHS ? g_strdup(info->hesiodLHS) : NULL;

        ret->ldapServer = info->ldapServer ? g_strdup(info->ldapServer) : NULL;
        ret->ldapBaseDN = info->ldapBaseDN? g_strdup(info->ldapBaseDN) : NULL;

        ret->kerberosRealm =
	info->kerberosRealm ? g_strdup(info->kerberosRealm) : NULL;

        ret->kerberosKDC =
	info->kerberosKDC ? g_strdup(info->kerberosKDC) : NULL;

        ret->kerberosAdminServer =
	info->kerberosAdminServer ? g_strdup(info->kerberosAdminServer) : NULL;

        ret->nisServer = info->nisServer ? g_strdup(info->nisServer) : NULL;
        ret->nisDomain = info->nisDomain ? g_strdup(info->nisDomain) : NULL;

	return ret;
}

static gboolean
non_empty(const char *string)
{
	return (string != NULL) && (strlen(string) > 0);
}
static gboolean
is_empty(const char *string)
{
	return (string == NULL) || (strlen(string) == 0);
}

gboolean
authInfoWriteHesiod(struct authInfoType *info)
{
	shvarFile *sv = NULL;

	if((sv = svNewFile(SYSCONFDIR "/hesiod.conf")) == NULL) {
		sv = svCreateFile(SYSCONFDIR "/hesiod.conf");
	}

	if(sv == NULL) {
		return FALSE;
	}

	if(info->hesiodLHS != NULL) {
		if(strlen(info->hesiodLHS) == 0) {
			g_free(info->hesiodLHS);
			info->hesiodLHS = NULL;
		}
	}
	svSetValue(sv, "lhs", info->hesiodLHS);
	if(info->hesiodRHS != NULL) {
		if(strlen(info->hesiodRHS) == 0) {
			g_free(info->hesiodRHS);
			info->hesiodRHS = NULL;
		}
	}
	svSetValue(sv, "rhs", info->hesiodRHS);

	svWriteFile(sv, 0644);
	svCloseFile(sv);

	return TRUE;
}

/* Write NIS setup to /etc/yp.conf. */
gboolean
authInfoWriteNIS(struct authInfoType *info)
{
	char *ibuf = NULL, *obuf = NULL, *p, *q;
	int fd, l;
	struct stat st;
	struct flock lock;
	gboolean written = FALSE;

	fd = open(SYSCONFDIR "/yp.conf", O_RDWR | O_CREAT, 0644);
	if(fd == -1) {
		return FALSE;
	}

	memset(&lock, 0, sizeof(lock));
	lock.l_type = F_WRLCK;
	if(fcntl(fd, F_SETLKW, &lock) == -1) {
		return FALSE;
	}

	if(fstat(fd, &st) == -1) {
		return FALSE;
	}

	/* Read in the old file. */
	ibuf = g_malloc0(st.st_size + 1);
	read(fd, ibuf, st.st_size);

	/* Determine the maximum length of the new file. */
	l = strlen(" domain ") + strlen(" broadcast ");
	l += info->nisDomain ? strlen(info->nisDomain) : 0;
	l += info->nisServer ? strlen(info->nisServer) : 0;
	obuf = g_malloc0(st.st_size + 1 + l);

	p = ibuf;
	while(*p != '\0') {
		/* Isolate a single line. */
		for(q = p; (*q != '\0') && (*q != '\n'); q++);
		if(*q != '\0') q++;

		/* If it's a 'domain' line, insert ours instead. */
		if(strncmp("domain", p, 6) == 0) {
			if(!written)
			if(non_empty(info->nisDomain)) {
				strcat(obuf, "domain ");
				strcat(obuf, info->nisDomain);
				/* Take an empty server name to mean that we
				 * want to use broadcast. */
				if(non_empty(info->nisServer)) {
					strcat(obuf, " server ");
					if(strchr(info->nisServer, ',')) {
						char *q;
						q = strchr(info->nisServer, ',');
						strncat(obuf, info->nisServer,
							q - info->nisServer);
					} else {
						strcat(obuf, info->nisServer);
					}
				} else {
					strcat(obuf, " broadcast");
				}
				strcat(obuf, "\n");

				if(non_empty(info->nisServer))
				if(strchr(info->nisServer, ',')) {
					p = strchr(info->nisServer, ',') + 1;
					while(strchr(p, ',')) {
						char *q;
						q = strchr(p, ',');
						strcat(obuf, "ypserver ");
						strncat(obuf, p, q - p);
						strcat(obuf, "\n");
						p = q + 1;
					}
					strcat(obuf, "ypserver ");
					strcat(obuf, p);
					strcat(obuf, "\n");
				}

				written = TRUE;
			}
		} else

		/* If it's a 'ypserver' line, insert ours instead. */
		if(strncmp("ypserver", p, 8) == 0) {
			if(!written)
			if(is_empty(info->nisDomain))
			if(non_empty(info->nisServer)) {
				char *p = info->nisServer;
				while(strchr(p, ',')) {
					char *q;
					q = strchr(p, ',');
					strcat(obuf, "ypserver ");
					strncat(obuf, p, q - p);
					strcat(obuf, "\n");
					p = q + 1;
				}
				strcat(obuf, "ypserver ");
				strcat(obuf, p);
				strcat(obuf, "\n");
				written = TRUE;
			}
		} else

		/* Otherwise, just copy the current line out. */
		strncat(obuf, p, q - p);
		p = q;
	}

	/* If we haven't encountered a domain line yet... */
	if(!written) {
		if(non_empty(info->nisDomain)) {
			strcat(obuf, "domain ");
			strcat(obuf, info->nisDomain);
			if(non_empty(info->nisServer)) {
				strcat(obuf, " server ");
				strcat(obuf, info->nisServer);
			} else {
				strcat(obuf, " broadcast");
			}
			strcat(obuf, "\n");
		} else {
			if(non_empty(info->nisServer)) {
				strcat(obuf, "ypserver ");
				strcat(obuf, info->nisServer);
				strcat(obuf, "\n");
			}
		}
	}

	/* Write it out and close it. */
	ftruncate(fd, 0);
	lseek(fd, 0, SEEK_SET);
	write(fd, obuf, strlen(obuf));
	close(fd);

	g_free(ibuf);
	g_free(obuf);

	return TRUE;
}

/* Write LDAP setup to an ldap.conf using host and base as keys. */
static gboolean
authInfoWriteLDAP2(struct authInfoType *info, const char *filename,
		   const char *host, const char *base, gboolean writessl)
{
	char *ibuf = NULL, *obuf = NULL, *p, *q;
	int fd, l;
	struct stat st;
	struct flock lock;
	gboolean wrotebasedn = FALSE, wroteserver = FALSE, wrotessl = FALSE;

	fd = open(filename, O_RDWR | O_CREAT, 0644);
	if(fd == -1) {
		return FALSE;
	}

	memset(&lock, 0, sizeof(lock));
	lock.l_type = F_WRLCK;
	if(fcntl(fd, F_SETLKW, &lock) == -1) {
		return FALSE;
	}

	if(fstat(fd, &st) == -1) {
		return FALSE;
	}

	/* Read in the old file. */
	ibuf = g_malloc0(st.st_size + 1);
	read(fd, ibuf, st.st_size);

	/* Determine the maximum length of the new file. */
	l = strlen(host) + 2 + strlen(base) + 2;
	l += info->ldapBaseDN ? strlen(info->ldapBaseDN) : 0;
	l += info->ldapServer ? strlen(info->ldapServer) : 0;
	l += strlen("ssl start_tls\n");
	obuf = g_malloc0((st.st_size + 1 + l) * 2);

	p = ibuf;
	while(*p != '\0') {
		/* Isolate a single line. */
		for(q = p; (*q != '\0') && (*q != '\n'); q++);
		if(*q != '\0') q++;

		/* If it's a 'host' line, insert ours instead. */
		if(strncmp(host, p, 4) == 0) {
			if(!wroteserver)
			if(non_empty(info->ldapServer)) {
				strcat(obuf, host);
				strcat(obuf, " ");
				strcat(obuf, info->ldapServer);
				strcat(obuf, "\n");
				wroteserver = TRUE;
			}
		} else

		/* If it's a 'base' line, insert ours instead. */
		if(strncmp(base, p, 4) == 0) {
			if(!wrotebasedn)
			if(non_empty(info->ldapBaseDN)) {
				strcat(obuf, base);
				strcat(obuf, " ");
				strcat(obuf, info->ldapBaseDN);
				strcat(obuf, "\n");
				wrotebasedn = TRUE;
			}
		} else

		/* If it's an 'ssl' line, insert ours instead. */
		if(writessl && (strncmp("ssl", p, 3)) == 0) {
			if(!wrotessl) {
				strcat(obuf, "ssl");
				strcat(obuf, " ");
				strcat(obuf,
				       info->enableLDAPS ? "start_tls" : "no");
				strcat(obuf, "\n");
				wrotessl = TRUE;
			}
		} else

		/* Otherwise, just copy the current line out. */
		strncat(obuf, p, q - p);
		p = q;
	}

	/* If we haven't encountered either of the config lines yet... */
	if(!wroteserver) {
		if(non_empty(info->ldapServer)) {
			strcat(obuf, host);
			strcat(obuf, " ");
			strcat(obuf, info->ldapServer);
			strcat(obuf, "\n");
		}
	}
	if(!wrotebasedn) {
		if(non_empty(info->ldapBaseDN)) {
			strcat(obuf, base);
			strcat(obuf, " ");
			strcat(obuf, info->ldapBaseDN);
			strcat(obuf, "\n");
		}
	}
	if(writessl && !wrotessl) {
		strcat(obuf, "ssl");
		strcat(obuf, " ");
		strcat(obuf, info->enableLDAPS ? "start_tls" : "no");
		strcat(obuf, "\n");
	}

	/* Write it out and close it. */
	ftruncate(fd, 0);
	lseek(fd, 0, SEEK_SET);
	write(fd, obuf, strlen(obuf));
	close(fd);

	g_free(ibuf);
	g_free(obuf);

	return TRUE;
}

gboolean
authInfoWriteLDAP(struct authInfoType *info)
{
	gboolean ret = TRUE;
	if(ret) {
		ret = authInfoWriteLDAP2(info, SYSCONFDIR "/ldap.conf",
					 "host", "base", TRUE);
	}
	if(ret) {
		/* Ignore errors here. */
		authInfoWriteLDAP2(info, SYSCONFDIR "/openldap/ldap.conf",
				   "HOST", "BASE", FALSE);
	}
	return ret;
}

static void
write_kdc(char *obuf, struct authInfoType *info)
{
	char *p = info->kerberosKDC;
	if(is_empty(p))
		return;
	while(strchr(p, ',')) {
		strcat(obuf, "  kdc = ");
		strncat(obuf, p, strchr(p, ',') - p);
		p = strchr(p, ',') + 1;
		strcat(obuf, "\n");
	}
	strcat(obuf, "  kdc = ");
		strcat(obuf, p);
	strcat(obuf, "\n");
}

static void
write_admin_server(char *obuf, struct authInfoType *info)
{
	char *p = info->kerberosAdminServer;
	if(is_empty(p))
		return;
	while(strchr(p, ',')) {
		strcat(obuf, "  admin_server = ");
		strncat(obuf, p, strchr(p, ',') - p);
		p = strchr(p, ',') + 1;
		strcat(obuf, "\n");
	}
	strcat(obuf, "  admin_server = ");
		strcat(obuf, p);
	strcat(obuf, "\n");
}

static void
write_realm(char *obuf, struct authInfoType *info)
{
	strcat(obuf, " ");
	strcat(obuf, info->kerberosRealm);
	strcat(obuf, " = {\n");
	write_kdc(obuf, info);
	write_admin_server(obuf, info);
	strcat(obuf, " }\n");
	strcat(obuf, "\n");
}

static int
comma_count(const char *string)
{
	int ret = 0;
	for(;string && (*string != '\0'); string++) {
		if(*string == ',') {
			ret++;
		}
	}
	return ret;
}

/* Write Kerberos 5 setup to /etc/krb5.conf. */
gboolean
authInfoWriteKerberos(struct authInfoType *info)
{
	char *ibuf = NULL, *obuf = NULL, *p, *q;
	int fd, l;
	struct stat st;
	struct flock lock;
	gboolean wrotekdc = FALSE, wroterealm = FALSE, wroteadmin = FALSE;
	gboolean wroterealms = FALSE, wrotelibdefaults = FALSE;
	gboolean wrotedefaultrealm = FALSE;
	char *section = NULL, *subsection = NULL;

	fd = open(SYSCONFDIR "/krb5.conf", O_RDWR | O_CREAT, 0644);
	if(fd == -1) {
		return FALSE;
	}

	memset(&lock, 0, sizeof(lock));
	lock.l_type = F_WRLCK;
	if(fcntl(fd, F_SETLKW, &lock) == -1) {
		return FALSE;
	}

	if(fstat(fd, &st) == -1) {
		return FALSE;
	}

	/* Read in the old file. */
	ibuf = g_malloc0(st.st_size + 1);
	read(fd, ibuf, st.st_size);

	/* Determine the maximum length of the new file. */
	l = strlen("[libdefaults]\n default_realm = \n\n[realm]\n  = { \n }\n");
	l += info->kerberosRealm ? strlen(info->kerberosRealm) * 2 : 1;
	l += strlen("  kdc = \n\n") * (comma_count(info->kerberosKDC) + 1);
	l += info->kerberosKDC ? strlen(info->kerberosKDC) * 2 : 0;
	l += strlen("  admin_server = \n\n") *
	     (comma_count(info->kerberosAdminServer) + 1);
	l += info->kerberosAdminServer ?
	     strlen(info->kerberosAdminServer) * 2 : 0;
	obuf = g_malloc0(st.st_size + 1 + l);

	p = ibuf;
	while(*p != '\0') {
		/* Isolate a single line. */
		char *l = p;
		for(q = p; (*q != '\0') && (*q != '\n'); q++);
		if(*q != '\0') q++;

		/* Skip over any whitespace. */
		for(;isspace(*p) && (*p != '\0') && (*p != '\n'); p++);

		/* If this is the "kdc" in our realm, replace it with
		 * the values we now have. */
		if((section != NULL) &&
		   (strcmp(section, "realms") == 0) &&
		   (subsection != NULL) &&
		   (non_empty(info->kerberosRealm)) &&
		   (strcmp(subsection, info->kerberosRealm) == 0) &&
		   (strncmp(p, "kdc", 3) == 0)) {
			if(!wrotekdc)
			if(info->kerberosKDC) {
				write_kdc(obuf, info);
				wrotekdc = TRUE;
			}
			p = q;
			continue;
		}

		/* If this is the "admin_server" in our realm, replace it with
		 * the values we now have. */
		if((section != NULL) &&
		   (strcmp(section, "realms") == 0) &&
		   (subsection != NULL) &&
		   (non_empty(info->kerberosRealm)) &&
		   (strcmp(subsection, info->kerberosRealm) == 0) &&
		   (strncmp(p, "admin_server", 3) == 0)) {
			if(!wroteadmin)
			if(info->kerberosAdminServer) {
				write_admin_server(obuf, info);
				wroteadmin = TRUE;
			}
			p = q;
			continue;
		}

		/* If we're in the realms section, but not in a realm, we'd
		 * better be looking at the beginning of one. */
		if((section != NULL) &&
		   (strcmp(section, "realms") == 0) &&
		   (subsection == NULL)) {
			char *q;
			for(q = p; !isspace(*q) && (*q != '\0'); q++);
			if(subsection) {
				g_free(subsection);
			}
			subsection = g_strndup(p, q - p);
			if(is_empty(subsection)) {
				g_free(subsection);
				subsection = NULL;
			} else {
				/* If this is the section for our realm, mark
				 * that. */
				if(strcmp(subsection,info->kerberosRealm) == 0){
					wroterealm = TRUE;
				}
			}
		}

		/* If it's the end of a subsection, mark that. */
		if((section != NULL) &&
		   (strcmp(section, "realms") == 0) &&
	   	   (subsection != NULL) &&
		   (strncmp(p, "}", 1) == 0)) {
			/* If it's the right section of realms, write out
			 * info we haven't already written. */
	   	   	if(non_empty(info->kerberosRealm) &&
	   	   	   (strcmp(subsection, info->kerberosRealm) == 0)) {
				if(!wrotekdc) {
					write_kdc(obuf, info);
				}
				if(!wroteadmin) {
					write_admin_server(obuf, info);
				}
			}
			if(subsection) {
				g_free(subsection);
			}
			subsection = NULL;
		}

		/* If we're in the libdefaults section, and this is the
		 * default_realm keyword, replace it with ours. */
		if((section != NULL) &&
		   (strcmp(section, "libdefaults") == 0) &&
		   (strncmp(p, "default_realm", 13) == 0)) {
			if(non_empty(info->kerberosRealm) &&
			   !wrotedefaultrealm) {
				strcat(obuf, " default_realm = ");
				strcat(obuf, info->kerberosRealm);
				strcat(obuf, "\n");
				wrotedefaultrealm = TRUE;
			}
			p = q;
			continue;
		}

		/* If it's the beginning of a section, record its name. */
		if(strncmp("[", p, 1) == 0) {
			char *q;
			p++;
			/* If the previous section was "realms", and we didn't
			 * see ours, write our realm out. */
			if((section != NULL) &&
			   (strcmp(section, "realms") == 0) &&
			   (non_empty(info->kerberosRealm)) &&
			   !wroterealm) {
				write_realm(obuf, info);
				wroterealm = TRUE;
			}
			/* If the previous section was "libdefaults", and we
			 * didn't see a "default_realm", write it out. */
			if((section != NULL) &&
			   (strcmp(section, "libdefaults") == 0) &&
			   (non_empty(info->kerberosRealm)) &&
			   !wrotedefaultrealm) {
				strcat(obuf, " default_realm = ");
				strcat(obuf, info->kerberosRealm);
				strcat(obuf, "\n");
				wrotedefaultrealm = TRUE;
			}
			for(q = p; ((*q != ']') && (*q != '\0')); q++);
			if(section) {
				g_free(section);
			}
			section = g_strndup(p, q - p);
			if(strcmp(section, "realms") == 0) {
				wroterealms = TRUE;
			}
			if(strcmp(section, "libdefaults") == 0) {
				wrotelibdefaults = TRUE;
			}
		}

		/* Otherwise, just copy the current line out. */
		strncat(obuf, l, q - l);
		p = q;
	}

	/* If we haven't encountered a libdefaults section yet... */
	if(!wrotelibdefaults && non_empty(info->kerberosRealm)) {
		strcat(obuf, "[libdefaults]\n");
		strcat(obuf, " default_realm = ");
		strcat(obuf, info->kerberosRealm);
		strcat(obuf, "\n\n");
	}

	/* If we haven't encountered a realms section yet... */
	if(!wroterealms && non_empty(info->kerberosRealm)) {
		strcat(obuf, "[realms]\n");
		write_realm(obuf, info);
	}

	/* Write it out and close it. */
	ftruncate(fd, 0);
	lseek(fd, 0, SEEK_SET);
	write(fd, obuf, strlen(obuf));
	close(fd);

	/* Clean up. */
	if(ibuf) {
		g_free(ibuf);
	}
	if(obuf) {
		g_free(obuf);
	}
	if(section) {
		g_free(section);
	}
	if(subsection) {
		g_free(subsection);
	}

	return TRUE;
}

/* Write NSS setup to /etc/nsswitch.conf. */
gboolean
authInfoWriteNSS(struct authInfoType *info)
{
	char *ibuf = NULL, *obuf = NULL, *p, *q;
	int fd, l;
	struct stat st;
	struct flock lock;
	char normal[LINE_MAX] = "", hosts[LINE_MAX] = "";
	gboolean wrotepasswd = FALSE, wrotegroup = FALSE, wroteshadow = FALSE,
		 wroteservices = FALSE, wroteprotocols = FALSE,
		 wrotenetgroup = FALSE, wroteautomount = FALSE,
		 wrotehosts = FALSE;

	fd = open(SYSCONFDIR "/nsswitch.conf", O_RDWR|O_CREAT, 0644);
	if(fd == -1) {
		return FALSE;
	}

	memset(&lock, 0, sizeof(lock));
	lock.l_type = F_WRLCK;
	if(fcntl(fd, F_SETLKW, &lock) == -1) {
		return FALSE;
	}

	if(fstat(fd, &st) == -1) {
		return FALSE;
	}

	/* Read in the old file. */
	ibuf = g_malloc0(st.st_size + 1);
	read(fd, ibuf, st.st_size);

	/* Determine the maximum length of the new file. */
	l = strlen("passwd:     \n") +
	    strlen("shadow:     \n") +
	    strlen("group:      \n") +
	    strlen("services:   \n") +
	    strlen("protocols:  \n") +
	    strlen("netgroup:   \n") +
	    strlen("automount:  \n") +
	    strlen("hosts:      \n");
	l += strlen(" files nisplus nis") * 8;
	l += strlen(" db") * 8;
	l += strlen(" files") * 8;
	l += strlen(" hesiod") * 8;
	l += strlen(" ldap") * 8;
	l += strlen(" nis") * 8;
	l += strlen(" nisplus") * 8;
	l += strlen(" dns");
	l += strlen(" winbind") * 8;
	obuf = g_malloc0(st.st_size + 1 + l);

	/* Determine what we want in that file for most of the databases. */
	if(info->enableDB) strcat(normal, " db");
	strcat(normal, " files");
	if(info->enableNIS3) strcat(normal, " nisplus");
	if(info->enableNIS) strcat(normal, " nis");
	if(info->enableLDAP) strcat(normal, " ldap");
	if(info->enableHesiod) strcat(normal, " hesiod");
	
	/* Hostnames we treat specially. */
	strcat(hosts, " files");
	if(info->enableNIS3) strcat(hosts, " nisplus");
	if(info->enableNIS) strcat(hosts, " nis");
	strcat(hosts, " dns");

	p = ibuf;
	while(*p != '\0') {
		/* Isolate a single line. */
		for(q = p; (*q != '\0') && (*q != '\n'); q++);
		if(*q != '\0') q++;

		/* If it's a 'passwd' line, insert ours instead. */
		if(strncmp("passwd:", p, 7) == 0) {
			if(!wrotepasswd) {
				strcat(obuf, "passwd:    ");
				strcat(obuf, normal);
				strcat(obuf, "\n");
				wrotepasswd = TRUE;
			}
		} else

		/* If it's a 'shadow' line, insert ours instead. */
		if(strncmp("shadow:", p, 7) == 0) {
			if(!wroteshadow) {
				strcat(obuf, "shadow:    ");
				strcat(obuf, normal);
				strcat(obuf, "\n");
				wroteshadow = TRUE;
			}
		} else

		/* If it's a 'group' line, insert ours instead. */
		if(strncmp("group:", p, 6) == 0) {
			if(!wrotegroup) {
				strcat(obuf, "group:     ");
				strcat(obuf, normal);
				strcat(obuf, "\n");
				wrotegroup = TRUE;
			}
		} else

		/* If it's a 'services' line, insert ours instead. */
		if(strncmp("services:", p, 9) == 0) {
			if(!wroteservices) {
				strcat(obuf, "services:  ");
				strcat(obuf, normal);
				strcat(obuf, "\n");
				wroteservices = TRUE;
			}
		} else

		/* If it's a 'protocols' line, insert ours instead. */
		if(strncmp("protocols:", p, 10) == 0) {
			if(!wroteprotocols) {
				strcat(obuf, "protocols: ");
				strcat(obuf, normal);
				strcat(obuf, "\n");
				wroteprotocols = TRUE;
			}
		} else

		/* If it's a 'netgroup' line, insert ours instead. */
		if(strncmp("netgroup:", p, 9) == 0) {
			if(!wrotenetgroup) {
				strcat(obuf, "netgroup:  ");
				strcat(obuf, normal);
				strcat(obuf, "\n");
				wrotenetgroup = TRUE;
			}
		} else

		/* If it's a 'automount' line, insert ours instead. */
		if(strncmp("automount:", p, 10) == 0) {
			if(!wroteautomount) {
				strcat(obuf, "automount: ");
				strcat(obuf, normal);
				strcat(obuf, "\n");
				wroteautomount = TRUE;
			}
		} else

		/* If it's a 'hosts' line, insert ours instead. */
		if(strncmp("hosts:", p, 6) == 0) {
			if(!wrotehosts) {
				strcat(obuf, "hosts:     ");
				strcat(obuf, hosts);
				strcat(obuf, "\n");
				wrotehosts = TRUE;
			}
		} else

		/* Otherwise, just copy the current line out. */
		strncat(obuf, p, q - p);
		p = q;
	}

	/* If we haven't encountered any of the config lines yet... */
	if(!wrotepasswd) {
		strcat(obuf, "passwd:    ");
		strcat(obuf, normal);
		strcat(obuf, "\n");
	}
	if(!wroteshadow) {
		strcat(obuf, "shadow:    ");
		strcat(obuf, normal);
		strcat(obuf, "\n");
	}
	if(!wrotegroup) {
		strcat(obuf, "group:     ");
		strcat(obuf, normal);
		strcat(obuf, "\n");
	}
	if(!wroteprotocols) {
		strcat(obuf, "protocols: ");
		strcat(obuf, normal);
		strcat(obuf, "\n");
	}
	if(!wroteservices) {
		strcat(obuf, "services:  ");
		strcat(obuf, normal);
		strcat(obuf, "\n");
	}
	if(!wrotenetgroup) {
		strcat(obuf, "netgroup:  ");
		strcat(obuf, normal);
		strcat(obuf, "\n");
	}
	if(!wroteautomount) {
		strcat(obuf, "automount: ");
		strcat(obuf, normal);
		strcat(obuf, "\n");
	}
	if(!wrotehosts) {
		strcat(obuf, "hosts:     ");
		strcat(obuf, hosts);
		strcat(obuf, "\n");
	}

	/* Write it out and close it. */
	ftruncate(fd, 0);
	lseek(fd, 0, SEEK_SET);
	write(fd, obuf, strlen(obuf));
	close(fd);

	g_free(ibuf);
	g_free(obuf);

	return TRUE;
}

/* Mandatory arguments for the various modules. */
#ifdef LOCAL_POLICIES
static const char *argv_local_all[] = {
	"service=" LOCAL_POLICY_NAME,
	NULL,
};
#endif

static const char *argv_unix_auth[] = {
	"likeauth",
	"nullok",
	NULL,
};
static const char *argv_unix_password[] = {
	"nullok",
	"use_authtok",
	NULL,
};

static const char *argv_krb5_auth[] = {
	"use_first_pass",
	NULL,
};

static const char *argv_krb5afs_auth[] = {
	"use_first_pass",
	"tokens",
	NULL,
};

static const char *argv_krb5_password[] = {
	"use_authtok",
	NULL,
};

static const char *argv_ldap_auth[] = {
	"use_first_pass",
	NULL,
};

static const char *argv_winbind_auth[] = {
	"use_first_pass",
	NULL,
};

static const char *argv_winbind_password[] = {
	"use_authtok",
	NULL,
};

static const char *argv_ldap_password[] = {
	"use_authtok",
	NULL,
};

static const char *argv_cracklib_password[] = {
	"retry=3",
	NULL,
};

/* Enumerations for PAM control flags and stack names. */
enum pam_function_type {
	auth,
	account,
	session,
	password,
};

/* The list of stacks, module flags, and arguments, if there are any. */
static struct {
	gboolean mandatory;
	enum pam_function_type stack;
	const char *logic;
	const char *name;
	const char **argv;
} standard_pam_modules[] = {
#ifdef LOCAL_POLICIES
	{FALSE, auth,		LOGIC_REQUIRED,
	 "stack",		argv_local_all},
#endif
	{TRUE,  auth,		LOGIC_REQUIRED,
	 "env",			NULL},
	{TRUE,  auth,		LOGIC_SUFFICIENT,
	 "unix",		argv_unix_auth},
	{FALSE, auth,		LOGIC_SUFFICIENT,
	 "krb5",		argv_krb5_auth},
	{FALSE, auth,		LOGIC_SUFFICIENT,
	 "krb5afs",		argv_krb5afs_auth},
	{FALSE, auth,		LOGIC_SUFFICIENT,
	 "ldap",		argv_ldap_auth},
	{FALSE, auth,		LOGIC_SUFFICIENT,
	 "winbind",		argv_winbind_auth},
	{TRUE,  auth,		LOGIC_REQUIRED,	
	 "deny",		NULL},

#ifdef LOCAL_POLICIES
	{FALSE, account,	LOGIC_REQUIRED,	
	 "stack",		argv_local_all},
#endif
	{TRUE,  account, 	LOGIC_REQUIRED,
	 "unix",		NULL},
	{FALSE, account,	LOGIC_IGNORE_UNKNOWN,
	 "ldap",		NULL},

#ifdef LOCAL_POLICIES
	{FALSE, password,	LOGIC_REQUIRED,
	 "stack",		argv_local_all},
#endif
	{TRUE,  password,	LOGIC_REQUIRED,
	 "cracklib",		argv_cracklib_password},
	{TRUE,  password,	LOGIC_SUFFICIENT,
	 "unix",		argv_unix_password},
	{FALSE, password,	LOGIC_SUFFICIENT,
	 "krb5",		argv_krb5_password},
	{FALSE, password,	LOGIC_SUFFICIENT,
	 "krb5afs",		argv_krb5_password},
	{FALSE, password,	LOGIC_SUFFICIENT,
	 "ldap",		argv_ldap_password},
	{FALSE, password,	LOGIC_SUFFICIENT,
	 "winbind",		argv_winbind_password},
	{TRUE,  password,	LOGIC_REQUIRED,
	 "deny",		NULL},

#ifdef LOCAL_POLICIES
	{FALSE, session,	LOGIC_REQUIRED,
	 "stack",		argv_local_all},
#endif
	{TRUE,  session,	LOGIC_REQUIRED,
	 "limits",		NULL},
	{TRUE,  session,	LOGIC_REQUIRED,
	 "unix",		NULL},
	{FALSE, session,	LOGIC_OPTIONAL,
	 "krb5",		NULL},
	{FALSE, session,	LOGIC_OPTIONAL,
	 "krb5afs",		NULL},
	{FALSE, session,	LOGIC_OPTIONAL,
	 "ldap",		NULL},
	{FALSE, session,	LOGIC_REQUIRED,
	 "winbind",		NULL},
};

static void
fmt_standard_pam_module(int i, char *obuf, struct authInfoType *info)
{
	char *stack;
	const char *logic = NULL;
	switch(standard_pam_modules[i].stack) {
		case auth:
			stack = "auth";
			break;
		case account:
			stack = "account";
			break;
		case session:
			stack = "session";
			break;
		case password:
			stack = "password";
			break;
		default:
			stack = NULL;
			break;
	}
	logic = standard_pam_modules[i].logic;
	if(non_empty(stack) && non_empty(logic)) {
		if(strlen(logic) > 0) {
			int j;
			char buf[LINE_MAX];
			memset(buf, '\0', sizeof(buf));
			snprintf(buf, sizeof(buf) - 1,
				 "%-12s%-13s %s/pam_%s.so", stack, logic,
				 AUTH_MODULE_DIR, standard_pam_modules[i].name);
			if(standard_pam_modules[i].argv != NULL) {
				for(j = 0;
				    non_empty(standard_pam_modules[i].argv[j]);
				    j++) {
					strncat(buf, " ",
						sizeof(buf) - 1 - strlen(buf));
					strncat(buf,
						standard_pam_modules[i].argv[j],
						sizeof(buf) - 1 - strlen(buf));
				}
			}
			if(strcmp(standard_pam_modules[i].name, "unix") == 0)
			if(stack != NULL) {
				if(strcmp(stack, "password") == 0) {
					if(info->enableMD5) {
						strncat(buf, " md5",
						sizeof(buf) - 1 - strlen(buf));
					}
					if(info->enableShadow) {
						strncat(buf, " shadow",
						sizeof(buf) - 1 - strlen(buf));
					}
					if(info->enableNIS) {
						strncat(buf, " nis",
						sizeof(buf) - 1 - strlen(buf));
					}
					if(info->enableBigCrypt) {
						strncat(buf, " bigcrypt",
						sizeof(buf) - 1 - strlen(buf));
					}
				}
				if((strcmp(stack, "account") == 0)) {
					if(info->brokenShadow) {
						strncat(buf, " broken_shadow",
						sizeof(buf) - 1 - strlen(buf));
					}
				}
			}
			strcat(obuf, buf);
		}
		strcat(obuf, "\n");
	}
}

/* Write PAM setup to the control file. */
gboolean authInfoWritePAM(struct authInfoType *authInfo)
{
	int i;
	char *obuf = NULL;
	int fd;
	struct flock lock;
	gboolean have_afs = FALSE;
	shvarFile *sv = NULL;

	fd = open(SYSCONFDIR "/pam.d/" AUTH_PAM_SERVICE, O_RDWR|O_CREAT, 0644);
	if(fd == -1) {
		return FALSE;
	}

	memset(&lock, 0, sizeof(lock));
	lock.l_type = F_WRLCK;
	if(fcntl(fd, F_SETLKW, &lock) == -1) {
		return FALSE;
	}

	obuf = g_malloc0(LINE_MAX *
		       	 (sizeof(standard_pam_modules) / 
		       	  sizeof(standard_pam_modules[0])));
	strcpy(obuf, "#%PAM-1.0\n");
	strcat(obuf, "# This file is auto-generated.\n");
      	strcat(obuf, "# User changes will be destroyed the next time "
		     "authconfig is run.\n");

	have_afs = (access("/afs", R_OK | X_OK) != -1);

	for(i = 0;
	    i < sizeof(standard_pam_modules) / sizeof(standard_pam_modules[0]);
	    i++) {
		if((i > 0) &&
		   (standard_pam_modules[i].stack !=
		    standard_pam_modules[i - 1].stack)) {
			strcat(obuf, "\n");
		}
		if(standard_pam_modules[i].mandatory ||
		   (authInfo->enableKerberos && !have_afs &&
		    (strcmp("krb5", standard_pam_modules[i].name) == 0)) ||
		   (authInfo->enableKerberos && have_afs &&
		    (strcmp("krb5afs", standard_pam_modules[i].name) == 0)) ||
#ifdef LOCAL_POLICIES
		   (authInfo->enableLocal &&
		    (strcmp("stack", standard_pam_modules[i].name) == 0)) ||
#endif
		   (authInfo->enableLDAPAuth &&
		    (strcmp("ldap", standard_pam_modules[i].name) == 0))) {
			fmt_standard_pam_module(i, obuf, authInfo);
		}
	}

	ftruncate(fd, 0);
	lseek(fd, 0, SEEK_SET);
	write(fd, obuf, strlen(obuf));
	g_free(obuf);
	close(fd);

	sv = svCreateFile(SYSCONFDIR "/sysconfig/authconfig");
	if(sv != NULL) {
		svSetValue(sv, "USEMD5",
			   authInfo->enableMD5 ? "yes" : "no");
		svSetValue(sv, "USEKERBEROS",
			   authInfo->enableKerberos ? "yes" : "no");
		svSetValue(sv, "USELDAPAUTH",
			   authInfo->enableLDAPAuth ? "yes" : "no");
		svWriteFile(sv, 0644);
		svCloseFile(sv);
	}

	return TRUE;
}

gboolean
authInfoWriteNetwork(struct authInfoType *info)
{
	shvarFile *sv = NULL;

	if((sv = svNewFile(SYSCONFDIR "/sysconfig/network")) == NULL) {
		sv = svCreateFile(SYSCONFDIR "/sysconfig/network");
	};
	if(sv == NULL) {
		return FALSE;
	}

	if(info->nisDomain != NULL) {
		if(strlen(info->nisDomain) == 0) {
			g_free(info->nisDomain);
			info->nisDomain = NULL;
		}
	}
	svSetValue(sv, "NISDOMAIN", info->nisDomain);
	svWriteFile(sv, 0644);
	svCloseFile(sv);
	return TRUE;
}

gboolean
authInfoWrite(struct authInfoType *authInfo)
{
	gboolean ret = TRUE;
	if(authInfo->enableKerberos)
		ret = ret && authInfoWriteKerberos(authInfo);
	if(authInfo->enableHesiod)
		ret = ret && authInfoWriteHesiod(authInfo);
	if(authInfo->enableNIS)
		ret = ret && authInfoWriteNIS(authInfo);
	if(authInfo->enableLDAP)
		ret = ret && authInfoWriteLDAP(authInfo);
	ret = ret && authInfoWriteNSS(authInfo);
	ret = ret && authInfoWritePAM(authInfo);
	ret = ret && authInfoWriteNetwork(authInfo);
	return ret;
}
